import unittest
from os.path import join

from tqdm import tqdm

from bulk_chain.api import CWD, iter_content
from bulk_chain.core.utils import dynamic_init
from utils import iter_test_jsonl_samples


class TestProviderBatching(unittest.TestCase):

    llm = dynamic_init(class_dir=join(CWD, ".."),
                       class_filepath="providers/transformers_flan_t5.py",
                       class_name="FlanT5")(model_name="nicolay-r/flan-t5-tsa-thor-base",
                                            max_new_tokens=128)

    def test_iter(self):
        input_dicts_it = iter_test_jsonl_samples()
        data_it = iter_content(input_dicts_it=input_dicts_it,
                               llm=self.llm,
                               batch_size=20,
                               return_batch=False,
                               schema="schema/default.json")

        for item in tqdm(data_it):
            print(item)


if __name__ == '__main__':
    unittest.main()



import json

import argparse
import logging
import sys

from source_iter.service_jsonl import JsonlService

from bulk_chain.api import init_llm
from bulk_chain.core.service_args import CmdArgsService
from bulk_chain.core.service_json import JsonService
from bulk_chain.core.service_llm import chat_with_lm
from bulk_chain.core.service_schema import SchemaService
from bulk_chain.core.utils import parse_filepath

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def iter_from_json(filepath):
    with open(filepath, "r") as f:
        content = json.load(f)
        for key, value in content.items():
            yield key, value


def iter_from_text_file(filepath):
    with open(filepath, "r") as f:
        yield filepath.split('.')[0], f.read()


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="LLM demo usage based on CoT schema")
    parser.add_argument('--adapter', dest='adapter', type=str, default=None)
    parser.add_argument('--attempts', dest='attempts', type=int, default=None)
    parser.add_argument('--src', dest='src', type=str, nargs="*", default=None)
    parser.add_argument('--schema', dest='schema', type=str, default=None,
                        help="Path to the JSON file that describes schema")
    parser.add_argument('--limit-prompt', dest="limit_prompt", type=int, default=None,
                        help="Optional trimming prompt by the specified amount of characters.")

    # Extract native arguments.
    native_args = CmdArgsService.extract_native_args(sys.argv, end_prefix="%%")
    args = parser.parse_args(args=native_args[1:])

    # Extract model-related arguments and Initialize Large Language Model.
    model_args = CmdArgsService.find_grouped_args(lst=sys.argv, starts_with="%%m", end_prefix="%%")
    model_args_dict = CmdArgsService.args_to_dict(model_args) | {"attempts": args.attempts}
    llm, llm_model_name = init_llm(adapter=args.adapter, **model_args_dict)

    # Setup schema.
    schema = SchemaService(json_data=JsonService.read(args.schema))
    schema_name = schema.src.get("name", None)
    if schema is not None:
        logger.info(f"Using schema: {schema_name}")

    output_providers = {
        "jsonl": lambda filepath, data_it, header:
            JsonlService.write(target=filepath,
                               data_it=map(lambda item: {key: item[i] for i, key in enumerate(header)}, data_it))
    }

    input_file_handlers = {
        "json": lambda filepath: iter_from_json(filepath),
        "txt": lambda filepath: iter_from_text_file(filepath)
    }

    # Input extension type defines the provider.
    if args.src is None:
        args.src = []
    if isinstance(args.src, str):
        args.src = [args.src]
    sources = [parse_filepath(s) for s in args.src]

    preset_dict = {}
    for fp, ext, _ in sources:
        for key, value in input_file_handlers[ext](fp):
            if key in preset_dict:
                raise Exception(f"While at handling {fp}: Key {key} is already registered!")
            preset_dict[key] = value

    # Launch Demo.
    chat_with_lm(llm, preset_dict=preset_dict, chain=schema.chain, model_name=llm_model_name,
                 line_width=120)



import random
import time
import logging


def setup_logger_behaviour(name: str) -> logging.Logger:
    root_handlers = logging.getLogger().handlers # gets root logger
    current_logger = logging.getLogger(name)   # gets current logger
    if not root_handlers:  # if root logger has no handlers then create streaming handeler only
        new_handler = logging.StreamHandler()
        new_handler.terminator = ""
        new_handler.setFormatter(logging.Formatter("%(message)s"))
        current_logger.addHandler(new_handler)
        current_logger.propagate = False
        current_logger.setLevel(logging.INFO)
        return current_logger

    # Remove  exixting Handlers from the current logger
    for handler in current_logger.handlers[:]:
        current_logger.removeHandler(handler)

    for handler_r in root_handlers:  # if root logger has handlers
        if type(handler_r) is logging.StreamHandler:  # if root logger has streaming handler
            new_handler = logging.StreamHandler() 
            new_handler.terminator = ""   # This will stop the printing in new line
            new_handler.setFormatter(logging.Formatter("%(message)s")) # This will set the format
            current_logger.addHandler(new_handler)
        elif type(handler_r) is logging.FileHandler:  # if root logger has file handler
            new_handler = logging.FileHandler( # create new file handler
                handler_r.baseFilename,        # with same filename and other properties
                handler_r.mode,
                handler_r.encoding,
                handler_r.delay,
                handler_r.errors,
            )
            new_handler.terminator = ""  # This will stop the printing in new line
            new_handler.setFormatter(logging.Formatter("%(message)s")) # This will set the format
            current_logger.addHandler(new_handler)
        else:
            continue
    current_logger.propagate = False  # Don't propagate to root logger
    return current_logger

# Configure the logger
logger =logging.getLogger(__name__)
class FakeStreamingDataGenerator:

    def stream_data(self):
        while True:
            data = random.randint(0, 100)
            yield data
            time.sleep(0.5)

# Example usage:
generator = FakeStreamingDataGenerator()
stream = generator.stream_data()

logger = setup_logger_behaviour(__name__) # call you set up function here
while True:
    chunk = next(stream)
    # Replacing print with logger
    logger.info(chunk)  # Best practice now


import os
from itertools import chain

from bulk_chain.core.llm_base import BaseLM
from bulk_chain.core.service_batch import BatchIterator, BatchService
from bulk_chain.core.service_data import DataService
from bulk_chain.core.service_dict import DictionaryService
from bulk_chain.core.service_json import JsonService
from bulk_chain.core.service_schema import SchemaService
from bulk_chain.core.utils import dynamic_init, find_by_prefix

INFER_MODES = {
    "default": lambda llm, prompt, limit_prompt=None: llm.ask_core(
        prompt[:limit_prompt] if limit_prompt is not None else prompt),
    "batch": lambda llm, batch, limit_prompt=None: llm.ask_core(
        DataService.limit_prompts(batch, limit=limit_prompt))
}


CWD = os.getcwd()


def _update_batch_content(c, batch, schema, infer_func):
    assert (isinstance(batch, list))
    assert (isinstance(c, str))

    if c in schema.p2r:
        for batch_item in batch:
            batch_item[c] = DataService.get_prompt_text(prompt=batch_item[c]["prompt"], data_dict=batch_item)
    if c in schema.r2p:
        p_column = schema.r2p[c]
        # This instruction takes a lot of time in a non-batching mode.
        BatchService.handle_param_as_batch(batch=batch,
                                           src_param=p_column,
                                           tgt_param=c,
                                           handle_func=lambda b: infer_func(b))


def _infer_batch(batch, schema, infer_func, cols=None):
    assert (isinstance(batch, list))
    assert (callable(infer_func))

    if len(batch) == 0:
        return batch

    if cols is None:
        first_item = batch[0]
        cols = first_item.keys() if cols is None else cols

    for c in cols:
        _update_batch_content(c=c, batch=batch, schema=schema, infer_func=infer_func)

    return batch


def iter_content(input_dicts_it, llm, schema, batch_size=1, return_batch=True, limit_prompt=None):
    """ This method represent Python API aimed at application of `llm` towards
        iterator of input_dicts via cache_target that refers to the SQLite using
        the given `schema`
    """
    assert (isinstance(llm, BaseLM))

    # Quick initialization of the schema.
    if isinstance(schema, str):
        schema = JsonService.read(schema)
    if isinstance(schema, dict):
        schema = SchemaService(json_data=schema)

    prompts_it = map(
        lambda data: DictionaryService.custom_update(src_dict=data, other_dict=schema.cot_args),
        input_dicts_it
    )

    content_it = (_infer_batch(batch=batch,
                               infer_func=lambda batch: INFER_MODES["batch"](llm, batch, limit_prompt),
                               schema=schema)
                  for batch in BatchIterator(prompts_it, batch_size=batch_size))

    yield from content_it if return_batch else chain.from_iterable(content_it)


def init_llm(adapter, **model_kwargs):
    """ This method perform dynamic initialization of LLM from third-party resource.
    """

    # List of the Supported models and their API wrappers.
    models_preset = {
        "dynamic": lambda: dynamic_init(class_dir=CWD, class_filepath=llm_model_name,
                                        class_name=llm_model_params)(**model_kwargs)
    }

    # Initialize LLM model.
    params = adapter.split(':')
    llm_model_type = params[0]
    llm_model_name = params[1] if len(params) > 1 else params[-1]
    llm_model_params = ':'.join(params[2:]) if len(params) > 2 else None
    llm = find_by_prefix(d=models_preset, key=llm_model_type)()

    return llm, llm_model_name



import importlib
import logging
import sys
from collections import Counter

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def find_by_prefix(d, key):
    """
        d: dict (str, val).
    """
    assert(isinstance(d, dict))
    assert(isinstance(key, str))

    # We first check the full match.
    for k, value in d.items():
        if k == key:
            return value

    # If we can't establish full match, then we seek by prefix.
    matches = []
    for k, value in d.items():
        if key.startswith(k):
            matches.append(k)

    if len(matches) > 1:
        raise Exception(f"There are multiple entries that are related to `{key}`: {matches}")
    if len(matches) == 0:
        raise Exception(f"No entries were found for {key}!")

    return d[matches[0]]


def iter_params(text):
    assert(isinstance(text, str))
    beg = 0
    while beg < len(text):
        try:
            pb = text.index('{', beg)
        except ValueError:
            break
        pe = text.index('}', beg+1)
        # Yield argument.
        yield text[pb+1:pe]
        beg = pe+1


def format_model_name(name):
    return name.replace("/", "_")


def parse_filepath(filepath, default_filepath=None, default_ext=None):
    """ This is an auxiliary function for handling sources and targets from cmd string.
    """
    if filepath is None:
        return default_filepath, default_ext, None
    info = filepath.split(":")
    filepath = info[0]
    meta = info[1] if len(info) > 1 else None
    ext = filepath.split('.')[-1] if default_ext is None else default_ext
    return filepath, ext, meta


def handle_table_name(name):
    return name.\
        replace('-', '_').\
        replace('.', "_")


def auto_import(name, is_class=False):
    """ Import from the external python packages.
    """
    def __get_module(comps_list):
        return importlib.import_module(".".join(comps_list))

    components = name.split('.')
    m = getattr(__get_module(components[:-1]), components[-1])

    return m() if is_class else m


def dynamic_init(class_dir, class_filepath, class_name=None):
    sys.path.append(class_dir)
    class_path_list = class_filepath.split('/')
    class_path_list[-1] = '.'.join(class_path_list[-1].split('.')[:-1])
    class_name = class_path_list[-1].title() if class_name is None else class_name
    class_path = ".".join(class_path_list + [class_name])
    logger.info(f"Dynamic loading for the file and class `{class_path}`")
    cls = auto_import(class_path, is_class=False)
    return cls


def optional_limit_iter(it_data, limit=None):
    counter = Counter()
    for data in it_data:
        counter["returned"] += 1
        if limit is not None and counter["returned"] > limit:
            break
        yield data



import unittest

from bulk_chain.core.utils import iter_params


class TestArgumentsSeeking(unittest.TestCase):

    def test(self):
        params = list(iter_params("X is a {x} and p is {text} and for {k}"))

        line = ",".join(["{{{x}}}".format(x=x) for x in params])
        print(line)
        d_params = {}
        for param in params:
            d_params[param] = 2
        print(d_params)

        z = line.format(**d_params)
        print(z)

        b = list(iter_params("X"))
        print(b)


if __name__ == '__main__':
    unittest.main()



import logging
import time

from bulk_chain.core.utils import format_model_name


class BaseLM(object):

    def __init__(self, name=None, attempts=None, delay_sec=1, enable_log=True,
                 support_batching=False, **kwargs):

        self.__name = name
        self.__attempts = 1 if attempts is None else attempts
        self.__delay_sec = delay_sec
        self.__support_batching = support_batching

        if enable_log:
            self.__logger = logging.getLogger(__name__)
            logging.basicConfig(level=logging.INFO)

    def ask_core(self, batch):

        for i in range(self.__attempts):
            try:
                if self.__support_batching:
                    # Launch in batch mode.
                    content = self.ask(batch)
                else:
                    # Launch in non-batch mode.
                    assert len(batch) == 1, "The LM does not support batching," \
                                            f" while size of the content is {len(batch)} which is not equal 1. " \
                                            f"Please enable batch-supporting or set required inference settings."
                    content = batch[0]

                response = self.ask(content)

                # Wrapping into batch the response in the case of non-batching mode.
                return response if self.__support_batching else [response]

            except Exception as e:
                if self.__logger is not None:
                    self.__logger.info("Unable to infer the result. Try {} out of {}.".format(i, self.__attempts))
                    self.__logger.info(e)
                time.sleep(self.__delay_sec)

        raise Exception("Can't infer")

    def ask(self, content):
        raise NotImplemented()

    def name(self):
        return format_model_name(self.__name)



import sys
import unittest

from bulk_chain.core.service_args import CmdArgsService


class TestCmdArgs(unittest.TestCase):

    def test(self):

        # Csv-related.
        csv_args = CmdArgsService.find_grouped_args(sys.argv, starts_with="%%csv", end_prefix="%%")
        print(csv_args)
        csv_args = CmdArgsService.args_to_dict(csv_args)
        print("csv\t", csv_args)

        # Model-related.
        m_args = CmdArgsService.find_grouped_args(sys.argv, starts_with="%%m", end_prefix="%%")
        m_args = CmdArgsService.args_to_dict(m_args)
        print("mod\t", m_args)

        # native.
        n_args = CmdArgsService.extract_native_args(sys.argv, end_prefix="%%")
        n_args = CmdArgsService.args_to_dict(n_args)
        print("nat\t", n_args)


if __name__ == '__main__':
    unittest.main()



import unittest
from os.path import join

from bulk_chain.api import iter_content, CWD
from bulk_chain.core.utils import dynamic_init
from bulk_chain.infer import iter_content_cached


class TestAPI(unittest.TestCase):

    llm = dynamic_init(class_dir=join(CWD, ".."),
                       class_filepath="providers/replicate_104.py",
                       class_name="Replicate")(api_token="<API-KEY>",
                                               model_name="deepseek-ai/deepseek-r1")

    def it_data(self, n):
        for i in range(n):
            yield {"ind": i, "text": "X invent sanctions against Y", "entity": "X"}

    def test_iter_cached(self):
        data_it = iter_content_cached(input_dicts_it=self.it_data(20),
                                      llm=self.llm,
                                      schema="../schema/default.json",
                                      # Cache-related extra parameters.
                                      cache_target="out.sqlite:content",
                                      id_column_name="ind")

        for data in data_it:
            print(data)

    def test_iter(self):
        data_it = iter_content(input_dicts_it=self.it_data(20),
                               llm=self.llm,
                               batch_size=1,
                               return_batch=True,
                               schema="../schema/default.json")

        for data in data_it:
            print(data)


if __name__ == '__main__':
    unittest.main()



from bulk_chain.core.llm_base import BaseLM
from bulk_chain.core.service_data import DataService
from bulk_chain.core.utils import iter_params
from bulk_chain.core.utils_logger import StreamedLogger


def pad_str(text, pad):
    return text.rjust(len(text) + pad, ' ')


def text_wrap(content, width, handle_line=lambda l: l):
    lines = []
    for text in content.split('\n'):
        for i in range(0, len(text), width):
            line = handle_line(text[i:i + width])
            lines.append(line)
    return '\n'.join(lines)


def nice_output(text, width, pad=4, remove_new_line=False):
    short_text = text.replace("\n", "") if remove_new_line else text
    return text_wrap(content=short_text, width=width, handle_line=lambda line: pad_str(line, pad=pad))


def chat_with_lm(lm, preset_dict=None, chain=None, model_name=None, line_width=80, pad=0):
    assert (isinstance(lm, BaseLM))
    assert (isinstance(chain, list))
    assert (isinstance(model_name, str) or model_name is None)

    preset_dict = {} if preset_dict is None else preset_dict

    streamed_logger = StreamedLogger(__name__)

    do_exit = False
    model_name = model_name if model_name is not None else "agent"

    while not do_exit:

        streamed_logger.info("----------------")
        streamed_logger.info("\n")

        # Launching the CoT engine loop.
        data_dict = {} | preset_dict
        for chain_ind, prompt_args in enumerate(chain):

            # Processing the prompt.
            prompt = prompt_args["prompt"]

            # Filling necessary parameters.
            user_informed = False
            field_names = list(iter_params(prompt))
            for ind, f_name in enumerate(field_names):

                if f_name in data_dict:
                    continue

                user_input = input(f"Enter your prompt for `{f_name}` ({ind+1}/{len(field_names)}) "
                                   f"(or 'exit' to quit): ")
                user_informed = True

                if user_input.lower() == 'exit':
                    do_exit = True
                    break

                data_dict[f_name] = user_input

            if do_exit:
                break

            # In the case of the initial interaction with the chain.
            # we make sure that aware user for starting interaction.
            if chain_ind == 0 and not user_informed:
                user_input = input(f"Enter to continue (or 'exit' to quit) ...")
                if user_input.lower() == 'exit':
                    do_exit = True

            # Finally asking LLM.
            DataService.compose_prompt_text(prompt=prompt, data_dict=data_dict, field_names=field_names)
            actual_prompt = DataService.get_prompt_text(prompt=prompt, data_dict=data_dict)

            # Returning meta information, passed to LLM.
            streamed_logger.info(pad_str(f"{model_name} (ask [{chain_ind+1}/{len(chain)}]) ->", pad=pad))
            streamed_logger.info("\n")
            streamed_logger.info(nice_output(actual_prompt, pad=pad, remove_new_line=True, width=line_width))
            streamed_logger.info("\n\n")

            # Response.
            response = lm.ask_core(batch=[actual_prompt])[0]
            streamed_logger.info(pad_str(f"{model_name} (resp [{chain_ind+1}/{len(chain)}])->", pad=pad))
            streamed_logger.info("\n")
            if isinstance(response, str):
                streamed_logger.info(nice_output(response, pad=pad, remove_new_line=False, width=line_width))
                buffer = [response]
            else:
                buffer = []
                for chunk in response:
                    streamed_logger.info(chunk)
                    buffer.append(str(chunk))

            streamed_logger.info("\n\n")

            # Collecting the answer for the next turn.
            data_dict[prompt_args["out"]] = "".join(buffer)



from os.path import join, basename

import argparse
import logging
import sys

from source_iter.service_csv import CsvService
from source_iter.service_jsonl import JsonlService
from source_iter.service_sqlite import SQLite3Service
from tqdm import tqdm

from bulk_chain.api import INFER_MODES, _infer_batch, CWD, init_llm
from bulk_chain.core.llm_base import BaseLM
from bulk_chain.core.service_args import CmdArgsService
from bulk_chain.core.service_dict import DictionaryService
from bulk_chain.core.service_json import JsonService
from bulk_chain.core.service_schema import SchemaService
from bulk_chain.core.utils import handle_table_name, optional_limit_iter, parse_filepath

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

WRITER_PROVIDERS = {
    "sqlite": lambda filepath, table_name, data_it, infer_data_func, **kwargs: SQLite3Service.write(
        data_it=data_it, target=filepath, table_name=table_name, data2col_func=infer_data_func,
        skip_existed=True, **kwargs)
}

READER_PROVIDERS = {
    "sqlite": lambda filepath, table_name: SQLite3Service.read(filepath, table=table_name)
}


def iter_content_cached(input_dicts_it, llm, schema, cache_target, limit_prompt=None, **cache_kwargs):
    assert (isinstance(llm, BaseLM))
    assert (isinstance(cache_target, str))

    # Quick initialization of the schema.
    if isinstance(schema, str):
        schema = JsonService.read(schema)
    if isinstance(schema, dict):
        schema = SchemaService(json_data=schema)

    # Iterator of the queries.
    prompts_it = map(
        lambda data: DictionaryService.custom_update(src_dict=data, other_dict=schema.cot_args),
        input_dicts_it
    )

    # Parse target.
    cache_filepath, _, cache_table = parse_filepath(filepath=cache_target)

    # Perform caching first.
    WRITER_PROVIDERS["sqlite"](
        filepath=cache_filepath, table_name=cache_table,
        data_it=tqdm(prompts_it, desc="Iter content"),
        infer_data_func=lambda c, prompt: _infer_batch(
            batch=[prompt], cols=[c],
            infer_func=lambda batch: INFER_MODES["default"](llm, batch, limit_prompt),
            schema=schema)[0][c],
        **cache_kwargs)

    # Then retrieve data.
    return READER_PROVIDERS["sqlite"](filepath=cache_filepath, table_name=cache_table)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Infer Instruct LLM inference based on CoT schema")
    parser.add_argument('--adapter', dest='adapter', type=str, default=None)
    parser.add_argument('--attempts', dest='attempts', type=int, default=None)
    parser.add_argument('--id-col', dest='id_col', type=str, default="uid")
    parser.add_argument('--src', dest='src', type=str, nargs="?", default=None)
    parser.add_argument('--schema', dest='schema', type=str, default=None,
                        help="Path to the JSON file that describes schema")
    parser.add_argument('--to', dest='to', type=str, default=None, choices=["csv", "sqlite"])
    parser.add_argument('--output', dest='output', type=str, default=None)
    parser.add_argument('--limit', dest='limit', type=int, default=None,
                        help="Limit amount of source texts for prompting.")
    parser.add_argument('--limit-prompt', dest="limit_prompt", type=int, default=None,
                        help="Optional trimming prompt by the specified amount of characters.")

    # Extract native arguments.
    native_args = CmdArgsService.extract_native_args(sys.argv, end_prefix="%%")
    args = parser.parse_args(args=native_args[1:])

    # Extract csv-related arguments.
    csv_args = CmdArgsService.find_grouped_args(lst=sys.argv, starts_with="%%csv", end_prefix="%%")
    csv_args_dict = CmdArgsService.args_to_dict(csv_args)

    # Extract model-related arguments and Initialize Large Language Model.
    model_args = CmdArgsService.find_grouped_args(lst=sys.argv, starts_with="%%m", end_prefix="%%")
    model_args_dict = CmdArgsService.args_to_dict(model_args) | {"attempts": args.attempts}
    llm, llm_model_name = init_llm(adapter=args.adapter, **model_args_dict)

    # Setup schema.
    schema = SchemaService(json_data=JsonService.read(args.schema))
    schema_name = schema.src.get("name", None)
    if schema is not None:
        logger.info(f"Using schema: {schema_name}")

    input_providers = {
        "csv": lambda filepath: CsvService.read(src=filepath, row_id_key=args.id_col,
                                                as_dict=True, skip_header=True,
                                                delimiter=csv_args_dict.get("delimiter", ","),
                                                escapechar=csv_args_dict.get("escapechar", None)),
        "tsv": lambda filepath: CsvService.read(src=filepath, row_id_key=args.id_col,
                                                as_dict=True, skip_header=True,
                                                delimiter=csv_args_dict.get("delimiter", "\t"),
                                                escapechar=csv_args_dict.get("escapechar", None)),
        "jsonl": lambda filepath: JsonlService.read(src=filepath, row_id_key=args.id_col)
    }

    output_providers = {
        "csv": lambda filepath, data_it, header: CsvService.write(target=filepath,
                                                                  data_it=data_it, header=header,
                                                                  delimiter=csv_args_dict.get("delimiter", ","),
                                                                  escapechar=csv_args_dict.get("escapechar", None),
                                                                  it_type=None),
        "tsv": lambda filepath, data_it, header: CsvService.write(target=filepath,
                                                                  data_it=data_it, header=header,
                                                                  delimiter=csv_args_dict.get("delimiter", "\t"),
                                                                  escapechar=csv_args_dict.get("escapechar", None),
                                                                  it_type=None),
        "jsonl": lambda filepath, data_it, header:
        JsonlService.write(target=filepath,
                           data_it=map(lambda item: {key: item[i] for i, key in enumerate(header)}, data_it))
    }

    # Setup output.
    args.output = args.output.format(model=llm.name()) if args.output is not None else args.output
    tgt_filepath, tgt_ext, tgt_meta = parse_filepath(args.output, default_ext=args.to)

    # We do not support multiple files for other modes.
    src_filepath, src_ext, src_meta = parse_filepath(args.src)

    def default_output_file_template(ext):
        # This is a default template for output files to be generated.
        return "".join(["_".join([join(CWD, basename(src_filepath)), llm.name(), schema_name]), ext])

    # Setup cache target as well as the related table.
    cache_filepath = default_output_file_template(".sqlite") if tgt_filepath is None else tgt_filepath
    cache_table = handle_table_name(tgt_meta if tgt_meta is not None else "contents")

    # This is a content that we extracted via input provider.
    it_data = input_providers[src_ext](src_filepath)

    data_it = iter_content_cached(input_dicts_it=optional_limit_iter(it_data=it_data, limit=args.limit),
                                  limit_prompt=args.limit_prompt,
                                  schema=schema,
                                  llm=llm,
                                  id_column_name=args.id_col,
                                  cache_target=":".join([cache_filepath, cache_table]))

    # Setup output target
    tgt_ext = src_ext if tgt_ext is None else tgt_ext
    output_target = default_output_file_template(f".{tgt_ext}") if tgt_filepath is None else tgt_filepath

    # Perform output writing process.
    output_providers[tgt_ext](filepath=output_target,
                              data_it=data_it,
                              header=SQLite3Service.read_columns(target=cache_filepath, table=cache_table))



class SchemaService(object):

    def __init__(self, json_data):
        self.src = json_data
        self.r2p, self.p2r, self.cot_args, self.chain = SchemaService.__init_schema(prompts=json_data["schema"])

    @classmethod
    def from_prompt(cls, prompt):
        prompt_schema = {"schema": [{"prompt": prompt, "out": "response", "in": "prompt"}]}
        return cls(prompt_schema)

    @staticmethod
    def __init_schema(prompts):

        schema_args = {}
        schema_r2p = {}
        schema_p2r = {}
        chain = []

        for prompt in prompts:
            r_col_name = prompt["out"]
            p_col_name = r_col_name + "_prompt" if "in" not in prompt else prompt["in"]

            assert r_col_name not in schema_r2p, f"`{r_col_name}` has been already declared!"
            assert p_col_name not in schema_p2r, f"`{p_col_name}` has been already declared!"

            schema_r2p[r_col_name] = p_col_name
            schema_p2r[p_col_name] = r_col_name
            schema_args[p_col_name] = prompt
            schema_args[r_col_name] = None
            chain.append(prompt)

        return schema_r2p, schema_p2r, schema_args, chain



class DictionaryService:

    @staticmethod
    def custom_update(src_dict, other_dict):
        for k, v in other_dict.items():
            if k in src_dict:
                raise Exception(f"The key `{k}` is already defined in both dicts with values: "
                                f"`{src_dict[k]}` (src) and `{v}` (other)")
            src_dict[k] = v
        return src_dict



class CmdArgsService:

    @staticmethod
    def autocast(v):
        for t in [int, float, str]:
            try:
                return t(v)
            except:
                pass

    @staticmethod
    def iter_arguments(lst):

        def __release():

            # We use the True value by default to treat the related parameter as flag.
            if len(buf) == 0:
                buf.append(True)

            return key, buf if len(buf) > 1 else buf[0]

        key = None
        buf = []
        for a in lst:
            if a.startswith('--'):
                # release
                if key is not None:
                    yield __release()
                # set new key and empty buf
                key = a[2:]
                buf = []
            else:
                # append argument into buffer.
                buf.append(a)

        # Sharing the remaining params.
        if key is not None:
            yield __release()

    @staticmethod
    def __find_suffix_ind(lst, idx_from, end_prefix):
        for i in range(idx_from, len(lst)):
            if lst[i].startswith(end_prefix):
                return i
        return len(lst)

    @staticmethod
    def extract_native_args(lst, end_prefix):
        return lst[:CmdArgsService.__find_suffix_ind(lst, idx_from=0, end_prefix=end_prefix)]

    @staticmethod
    def find_grouped_args(lst, starts_with, end_prefix):
        """Slices a list in two, cutting on index matching "sep"
        """

        # Checking the presence of starts_with.
        # We have to return empty content in the case of absence starts_with in the lst.
        if starts_with not in lst:
            return []

        # Assigning start index.
        idx_from = lst.index(starts_with) + 1

        # Assigning end index.
        idx_to = CmdArgsService.__find_suffix_ind(lst, idx_from=idx_from, end_prefix=end_prefix)

        return lst[idx_from:idx_to]

    @staticmethod
    def args_to_dict(args):
        return {k: CmdArgsService.autocast(v) if not isinstance(v, list) else v
                for k, v in CmdArgsService.iter_arguments(args)} if args is not None else {}



import logging


def StreamedLogger(name: str) -> logging.Logger:
    """ https://medium.com/@r.das699/optimizing-logging-practices-for-streaming-data-in-python-521798e1ed82
    """
    root_handlers = logging.getLogger().handlers
    current_logger = logging.getLogger(name)
    if not root_handlers:
        new_handler = logging.StreamHandler()
        new_handler.terminator = ""
        new_handler.setFormatter(logging.Formatter("%(message)s"))
        current_logger.addHandler(new_handler)
        current_logger.propagate = False
        current_logger.setLevel(logging.INFO)
        return current_logger

    for handler in current_logger.handlers[:]:
        current_logger.removeHandler(handler)

    for handler_r in root_handlers:
        if type(handler_r) is logging.StreamHandler:
            new_handler = logging.StreamHandler()
            new_handler.terminator = ""
            new_handler.setFormatter(logging.Formatter("%(message)s"))
            current_logger.addHandler(new_handler)
        elif type(handler_r) is logging.FileHandler:
            new_handler = logging.FileHandler(
                handler_r.baseFilename,
                handler_r.mode,
                handler_r.encoding,
                handler_r.delay,
                handler_r.errors,
            )
            new_handler.terminator = ""  # This will stop the printing in new line
            new_handler.setFormatter(logging.Formatter("%(message)s"))
            current_logger.addHandler(new_handler)
        else:
            continue
    current_logger.propagate = False  # Don't propagate to root logger
    return current_logger


from setuptools import (
    setup,
    find_packages,
)


def get_requirements(filenames):
    r_total = []
    for filename in filenames:
        with open(filename) as f:
            r_local = f.read().splitlines()
            r_total.extend(r_local)
    return r_total


setup(
    name='bulk_chain',
    version='0.25.1',
    python_requires=">=3.6",
    description='A lightweight, no-strings-attached Chain-of-Thought framework for your LLM, '
                'ensuring reliable results for bulk input requests.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/nicolay-r/bulk-chain',
    author='Nicolay Rusnachenko',
    author_email='rusnicolay@gmail.com',
    license='MIT License',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords='natural language processing, '
             'chain-of-thought, '
             'reasoning',
    packages=find_packages(),
    install_requires=get_requirements(['dependencies.txt'])
)


from bulk_chain.core.utils import iter_params


class DataService(object):

    @staticmethod
    def compose_prompt_text(prompt, data_dict, field_names):
        assert(isinstance(data_dict, dict))
        fmt_d = {col_name: data_dict[col_name] for col_name in field_names}

        # Guarantee that items has correct type.
        for k, v in fmt_d.items():
            if not isinstance(v, str):
                Exception("'{k}' parameter is expected to be string, but received '{v}'")

        return prompt.format(**fmt_d)

    @staticmethod
    def get_prompt_text(prompt, data_dict, parse_fields_func=iter_params):
        field_names = list(parse_fields_func(prompt))
        return DataService.compose_prompt_text(
            prompt=prompt, data_dict=data_dict, field_names=field_names)

    @staticmethod
    def limit_prompts(prompts_list, limit=None):
        return [p[:limit] if limit is not None else p for p in prompts_list]



class BatchService(object):

    @staticmethod
    def handle_param_as_batch(batch, src_param, tgt_param, handle_func):
        assert (isinstance(batch, list))
        assert (isinstance(src_param, str))
        assert (callable(handle_func))

        _batch = [item[src_param] for item in batch]

        # Do handling for the batch.
        _handled_batch = handle_func(_batch)
        assert (isinstance(_handled_batch, list))

        # Apply changes.
        for i, item in enumerate(batch):
            item[tgt_param] = _handled_batch[i]


class BatchIterator:

    def __init__(self, data_iter, batch_size, end_value=None):
        assert(isinstance(batch_size, int) and batch_size > 0)
        assert(callable(end_value) or end_value is None)
        self.__data_iter = data_iter
        self.__index = 0
        self.__batch_size = batch_size
        self.__end_value = end_value

    def __iter__(self):
        return self

    def __next__(self):
        buffer = []
        while True:
            try:
                data = next(self.__data_iter)
            except StopIteration:
                break
            buffer.append(data)
            if len(buffer) == self.__batch_size:
                break

        if len(buffer) > 0:
            self.__index += 1
            return buffer

        if self.__end_value is None:
            raise StopIteration
        else:
            return self.__end_value()
